"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Relayer = exports.RegularHttpClient = exports.TorHttpClient = exports.TorProvider = void 0;
// HTTP and proxy
const axios_1 = __importDefault(require("axios"));
const socks_proxy_agent_1 = require("socks-proxy-agent");
// Ethers
const ethers_1 = require("ethers");
const providers_1 = require("@ethersproject/providers");
// Local
const sdk_utils_1 = require("@tornado/sdk-utils");
const sdk_data_1 = require("@tornado/sdk-data");
// It seems that the default HttpProvider offered by the normal web3 package
// has some logic which either ignores the SocksProxyAgent or which falls back to
// using no proxy for some reason. In any case, the Tornado-modified one, originally
// modified by the Tornado Team or whoever else conributed, seems to properly error
// out when Tor is not running.
const HttpProvider = require('web3-providers-http');
/**
 * You can also set up a SOCKS5 I2P tunnel on some port and then use that instead. Meaning that this should be compatible with I2P.
 */
class TorProvider extends providers_1.Web3Provider {
    constructor(url, torOpts, network) {
        torOpts = torOpts ?? {};
        torOpts.rv = torOpts.rv ?? '102.0';
        const torPort = torOpts.port ?? 9050, headers = torOpts.headers ?? [
            {
                name: 'User-Agent',
                value: `Mozilla/5.0 (Windows NT 10.0; rv:${torOpts.rv}) Gecko/20100101 Firefox/${torOpts.rv}`
            }
        ];
        super(new HttpProvider(url, {
            // The h after socks5 means that DNS resolution is also done through Tor
            agent: { https: new socks_proxy_agent_1.SocksProxyAgent('socks5h://127.0.0.1:' + torPort) },
            // The XHR2 XMLHttpRequest assigns a Tor Browser header by itself.
            // But if in Browser we assign just in case.
            headers: typeof window !== 'undefined' ? headers : undefined,
            // 1 minute timeout, although not sure if this is behaving properly
            timeout: 60000
        }), network);
    }
}
exports.TorProvider = TorProvider;
// @ts-ignore
const TorHttpClient = function (opts) {
    const rv = opts?.rv ?? '102.0';
    return axios_1.default.create({
        headers: opts?.headers ?? {
            'User-Agent': `Mozilla/5.0 (Windows NT 10.0; rv:${rv}) Gecko/20100101 Firefox/${rv}`
        },
        httpsAgent: new socks_proxy_agent_1.SocksProxyAgent('socks5h://127.0.0.1:' + opts?.port ?? 9050),
        httpAgent: new socks_proxy_agent_1.SocksProxyAgent('socks5h://127.0.0.1:' + opts?.port ?? 9050),
        // 2 minute timeout
        timeout: 120000
    });
};
exports.TorHttpClient = TorHttpClient;
// @ts-ignore
const RegularHttpClient = function (opts) {
    return axios_1.default.create(opts);
};
exports.RegularHttpClient = RegularHttpClient;
class Relayer {
    url;
    httpClient;
    _fetched;
    _address;
    _version;
    _serviceFee;
    _miningFee;
    _status;
    _chainId;
    _prices;
    constructor(options, properties) {
        this.url = options.url;
        this.httpClient = options.httpClient;
        this._fetched = false;
        if (properties) {
            this._address = properties.address;
            this._version = properties.version;
            this._chainId = properties.chainId;
            this._serviceFee = properties.serviceFeePercent;
            this._miningFee = properties.miningFeePercent;
            this._status = properties.status;
            this._prices = properties.prices;
            this._fetched = true;
        }
    }
    // Setup
    /**
     * This function MUST be called to unlock the rest of the `Relayer` class functionality, as otherwise we don't have the property data necessary for all the logic we want.
     * @returns Fetched `RelayerProperties`.
     */
    async fetchProperties() {
        const properties = await this.httpClient
            .get(this.url + '/status')
            .catch((err) => {
            throw sdk_utils_1.ErrorUtils.ensureError(err);
        })
            .then((res) => res.data);
        if (Object.entries(properties).length === 0)
            throw sdk_utils_1.ErrorUtils.getError('Relayer.fetchProperties: Something went wrong with fetching properties from relayer endpoint.');
        this._address = properties['rewardAccount'];
        this._version = properties['version'];
        this._chainId = properties['netId'];
        this._serviceFee = properties['tornadoServiceFee'];
        this._miningFee = properties['miningFee'];
        this._status = properties['health']['status'];
        this._prices = Object.entries(properties['ethPrices']).reduce((map, entry) => map.set(entry[0], ethers_1.BigNumber.from(entry[1])), new Map());
        this._fetched = true;
        return {
            address: this._address,
            version: this._version,
            chainId: this._chainId,
            serviceFeePercent: this._serviceFee,
            miningFeePercent: this._miningFee,
            status: this._status,
            prices: this._prices
        };
    }
    _propertiesFetched(parentCallName) {
        if (!this._fetched)
            throw sdk_utils_1.ErrorUtils.getError(`Relayer.${parentCallName}: properties must be fetched first with \`fetchProperties\`.`);
    }
    // Getters
    get address() {
        this._propertiesFetched('address');
        return this._address;
    }
    get version() {
        this._propertiesFetched('version');
        return this._version;
    }
    get serviceFeePercent() {
        this._propertiesFetched('serviceFee');
        return this._serviceFee;
    }
    get miningFeePercent() {
        this._propertiesFetched('miningFee');
        return this._miningFee;
    }
    get status() {
        this._propertiesFetched('status');
        return this._status;
    }
    get chainId() {
        this._propertiesFetched('chainId');
        return this._chainId;
    }
    get prices() {
        this._propertiesFetched('prices');
        return this._prices;
    }
    get properties() {
        this._propertiesFetched('properties');
        return {
            address: this._address,
            version: this._version,
            chainId: this._chainId,
            serviceFeePercent: this._serviceFee,
            miningFeePercent: this._miningFee,
            status: this._status,
            prices: this._prices
        };
    }
    async getETHPurchasePrice(token) {
        return ethers_1.BigNumber.from(await this.httpClient
            .get(this.url + '/status')
            .catch((err) => {
            throw sdk_utils_1.ErrorUtils.ensureError(err);
        })
            .then((res) => ethers_1.BigNumber.from(res.data['ethPrices'][token])));
    }
    async handleWithdrawal(instanceAddress, proof) {
        const response = (await this.httpClient
            .post(this.url + '/v1/tornadoWithdraw', {
            contract: instanceAddress,
            proof: proof[0],
            args: proof.slice(1)
        })
            .catch(this._handleHTTPError));
        const { id } = response.data;
        let result = { success: false }, finished = false;
        while (!finished) {
            const statusResponse = (await this.httpClient
                .get(this.url + '/v1/jobs/' + id)
                .catch((err) => this._handleHTTPError(err, false)));
            if (statusResponse.status === 200) {
                const { txHash, status, _, failedReason } = statusResponse.data;
                if (status === 'FAILED') {
                    console.error(`\nRelayer.handleWithdrawal: withdrawal failed with reason: ${failedReason}\n`);
                    finished = true;
                }
                if (status === 'CONFIRMED') {
                    result.success = true;
                    result.txHash = txHash;
                    finished = true;
                }
            }
            await new Promise((resolve) => setTimeout(resolve, 3000));
        }
        return result;
    }
    _handleHTTPError(err, httpThrow = true) {
        err = sdk_utils_1.ErrorUtils.ensureError(err);
        if (err.response) {
            if (httpThrow)
                // @ts-expect-error
                throw sdk_utils_1.ErrorUtils.ensureError(err.response.data.error);
            // @ts-expect-error
            else
                console.error(err.response.data.error);
        }
        else if (err.request) {
            if (httpThrow)
                throw sdk_utils_1.ErrorUtils.ensureError(err.request);
            else
                console.error(err.request);
        }
        else
            throw err;
    }
    // Cache
    /**
     * Construct a new Relayer by reading relayer data from cache.
     */
    static async fromCache(options) {
        const cache = new sdk_data_1.Cache.Base('Relayers');
        // Error is ensured already
        const properties = await cache.get([options.url]).catch(() => {
            throw sdk_utils_1.ErrorUtils.getError(`Relayer.fromCache: relayer ${options.url} isn't stored in cache.`);
        });
        await cache.close().catch((err) => {
            throw sdk_utils_1.ErrorUtils.ensureError(err);
        });
        return new Relayer(options, properties);
    }
    /**
     * Cache relayer data into a PouchDB database in your cache folder. This will automatically fetch properties if they are not fetched.
     */
    async remember() {
        this._propertiesFetched('remember');
        const cache = new sdk_data_1.Cache.Base('Relayers');
        const doc = new sdk_data_1.Docs.Relayer(this.url, {
            address: this._address,
            version: this._version,
            chainId: this._chainId,
            serviceFeePercent: this._serviceFee,
            miningFeePercent: this._miningFee,
            status: this._status,
            prices: this._prices
        });
        await cache.db.put(doc).catch((err) => {
            throw sdk_utils_1.ErrorUtils.ensureError(err);
        });
        await cache.close().catch((err) => {
            throw sdk_utils_1.ErrorUtils.ensureError(err);
        });
    }
}
exports.Relayer = Relayer;
//# sourceMappingURL=index.js.map